from kratos import Generator, always_ff, verilog, posedge, always_comb
from kratos.util import reduce_add
import _kratos
import sqlite3
import tempfile
import os
import pytest
import json


def get_line_num(txt):
    with open(__file__) as f:
        lines = f.readlines()
    for i, line in enumerate(lines):
        if line.strip() == txt:
            return i + 1
    return 0


def test_db_dump():
    mod = Generator("mod", True)
    comb = mod.combinational()
    comb.add_stmt(mod.var("a", 1).assign(mod.var("b", 1)))

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        # hashing and generate verilog
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db)
        with open(debug_db) as f:
            db = json.load(f)
    assert db["generator"] == "kratos"
    assert db["top"] == mod.name
    mod_def = db["table"][0]
    assert mod_def["name"] == mod.name
    assert len(mod_def["scope"]) == 1
    comb = mod_def["scope"][0]
    assert comb["filename"] == __file__
    assert len(comb["scope"]) == 1
    assign = comb["scope"][0]
    ref_line = get_line_num('comb.add_stmt(mod.var("a", 1).assign(mod.var("b", 1)))')
    assert assign["line"] == ref_line
    var = assign["variable"]
    assert var["name"] == "a"


def test_debug_mock():
    # this is used for the runtime debugging
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)

            # ports
            self.in1 = self.input("in1", 16)
            self.in2 = self.input("in2", 16)
            self.out = self.output("out", 16)

            self.add_always(self.code)

        @always_comb
        def code(self):
            if self.in1 == 2:
                self.out = 2
            elif self.in1 == 1:
                self.out = 0
            elif self.in2 == 1:
                self.out = 1
            else:
                self.out = 3

    with tempfile.TemporaryDirectory() as temp:
        mod = Mod()
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                insert_verilator_info=True, debug_db_filename=debug_db)
        # make sure the verilator public is there
        with open(filename) as f:
            content = f.read()
        assert "verilator public" in content


def test_seq_debug():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            # ports
            self.in_ = self.input("in1", 1)
            self.clock("clk")
            for i in range(4):
                self.output("out{0}".format(i), 1)

            self.add_always(self.code1)
            self.add_always(self.code2)

        @always_comb
        def code1(self):
            if self.in_ == 0:
                self.ports.out0 = 0
                self.ports.out1 = 0
            else:
                self.ports.out0 = 1
                self.ports.out1 = 1

        @always_ff((posedge, "clk"))
        def code2(self):
            if self.in_ == 0:
                self.ports.out2 = 0
                self.ports.out3 = 0
            else:
                self.ports.out2 = 1
                self.ports.out3 = 1

    mod = Mod()
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        with open(debug_db) as f:
            db = json.load(f)
    # it should have two scopes
    mod_db = db["table"][0]
    assert mod_db["name"] == mod.name
    scopes = mod_db["scope"]
    assert len(scopes) == 2
    variables = {}
    for v in db["variables"]:
        variables[v["id"]] = v
    for i in range(2):
        code = scopes[i]
        decl = code["scope"][0]
        assert decl["type"] == "decl"
        var = db["variables"][-1]
        # this one is i, which should be compressed
        v_id = decl["variable"]
        assert v_id in variables
        var = variables[v_id]
        assert var["name"] == "i"
        assert var["value"] == "3"
        if_stmt = code["scope"][1]
        assert len(if_stmt["scope"]) == 2
        for stmt in if_stmt["scope"]:
            assert len(stmt["scope"]) == 2
            # two assignments
            for a in stmt["scope"]:
                assert a["type"] == "assign"
    # test out gen var
    assert len(mod_db["variables"]) == 1
    assert mod_db["variables"][0]["name"] == "self.in_"


def test_context():
    class Mod(Generator):
        def __init__(self, width):
            super().__init__("mod", debug=True)
            in_ = self.input("in", width)
            out = self.output("out", width)
            sel = self.input("sel", width)
            # test self variables
            self.out = out
            self.width = width

            @always_comb
            def code():
                if sel:
                    out = 0
                else:
                    for i in range(width):
                        out[i] = 1
            self.add_always(code)

    w = 4
    mod = Mod(w)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        with open(debug_db) as f:
            db = json.load(f)
        mod_db = db["table"][0]
        code = mod_db["scope"][-1]
        else_blk = code["scope"][-1]["scope"][-1]["scope"]
        assert len(else_blk) == (w * 2)
        values = set()
        for entry in else_blk:
            if entry["type"] == "decl" and entry["variable"]["name"] == "i":
                values.add(int(entry["variable"]["value"]))
        for i in range(w):
            assert i in values


def test_clock_interaction():
    mods = []
    num_child = 4
    for i in range(num_child):
        mod = Generator("mod", True)
        in_ = mod.input("in", 4)
        out_ = mod.output("out", 4)
        clk = mod.clock("clk")
        seq = mod.sequential((posedge, clk))
        seq.add_stmt(out_.assign(in_))
        mods.append(mod)
    parent = Generator("parent", True)
    clk = parent.clock("clk")
    in_ = parent.input("in", 4)
    out = parent.output("out", 4)
    for i, mod in enumerate(mods):
        parent.add_child("mod{0}".format(i), mod)
        parent.wire(mod.ports.clk, clk)
        if i == 0:
            continue
        parent.wire(mod.ports["in"], mods[i - 1].ports.out)
    parent.wire(mods[0].ports["in"], in_)
    parent.wire(out, mods[-1].ports.out)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(parent, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)


@pytest.mark.skip(reason="DPI-based assertion no longer supported")
def test_assert():
    from kratos import assert_
    mod = Generator("mod", True)
    in_ = mod.input("in", 1)
    out_ = mod.output("out", 1)

    @always_comb
    def code():
        # we introduce this bug on purpose
        out_ = in_ - 1
        assert_(out_ == in_)

    mod.add_always(code)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        with open(filename) as f:
            content = f.read()
            assert "assert (out == in) else" in content
        conn = sqlite3.connect(debug_db)
        c = conn.cursor()
        c.execute("SELECT * FROM breakpoint")
        lines = c.fetchall()
        assert len(lines) == 2
        # they are only one line apart
        assert abs(lines[0][3] - lines[1][3]) == 1
        conn.close()
    # once we remove the assertion, it should not be there
    _kratos.passes.remove_assertion(mod.internal_generator)
    src = verilog(mod)[0]["mod"]
    assert "assert" not in src


def test_wire():
    mod = Generator("mod", True)
    in_ = mod.input("in", 1)
    out_ = mod.output("out", 1)
    # context
    a = 1
    mod.wire(out_, in_)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        filename = os.path.join(temp, "test.sv")
        verilog(mod, filename=filename, insert_debug_info=True,
                debug_db_filename=debug_db)
        with open(debug_db) as f:
            db = json.load(f)
        assign = db["table"][0]["scope"][0]["scope"][-1]
        assert assign["type"] == "assign"
        assert assign["variable"]["name"] == "out"
        line = get_line_num("mod.wire(out_, in_)")
        assert assign["line"] == line


def test_multiple_definitions():
    def create_mod(idx):
        m = Generator("mod" + str(idx), True)
        in_ = m.input("in", 1)
        out = m.output("out", 1)
        comb = m.combinational()
        comb.add_stmt(out.assign(in_))
        return m
    mod = Generator("parent", True)
    input_ = mod.input("in", 1)
    output = mod.output("out", 1)
    mods = [create_mod(i) for i in range(2)]
    expr = None
    for i, m_ in enumerate(mods):
        mod.add_child("mod{0}".format(i), m_)
        mod.wire(input_, m_.ports["in"])
        if expr is None:
            expr = m_.ports["out"]
        else:
            expr = expr & m_.ports["out"]
    mod.wire(output, expr)

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db,
                optimize_passthrough=False)
        with open(debug_db) as f:
            db = json.load(f)
        assert len(db["table"]) == 3


def test_empty():
    from kratos.debug import dump_external_database
    mod = Generator("mod", True)
    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        dump_external_database([mod], "mod", debug_db)


def test_nested_scope():
    from kratos import clog2
    mod = Generator("FindHighestBit", True)
    width = 4
    data = mod.input("data", width)
    h_bit = mod.output("h_bit", clog2(width))
    done = mod.var("done", 1)

    @always_comb
    def find_bit():
        done = 0
        h_bit = 0
        for i in range(width):
            if ~done:
                if data[i]:
                    done = 1
                    h_bit = i
    mod.add_always(find_bit, label="block")
    verilog(mod, insert_debug_info=True)
    block = mod.get_marked_stmt("block")
    last_if = block[-1]
    for i in range(len(last_if.then_[-1].then_)):
        stmt = last_if.then_[-1].then_[i]
        context = stmt.scope_context
        if len(context) > 0:
            assert "i" in context
            is_var, var = context["i"]
            assert not is_var
            assert var == "3"


def test_array_packed():
    from kratos import PackedStruct
    mod = Generator("mod", True)
    aa = mod.var("a", 2, size=(2, 4), packed=True)
    struct = PackedStruct("s", [("read", 16, False),
                                ("data", 16, False)])
    ss = mod.var_packed("s", struct)
    mod.add_stmt(aa.assign(4))
    setattr(mod, "sss", ss)
    mod.add_stmt(ss["read"].assign(0))
    mod.add_stmt(ss["data"].assign(1))

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, debug_db_filename=debug_db, insert_debug_info=True)
        with open(debug_db) as f:
            db = json.load(f)
    mod_db = db["table"][0]
    mod_variables = mod_db["variables"]
    assert len(mod_variables) == 2
    assert mod_variables[0]["name"] == "self.sss.read"
    assert mod_variables[1]["name"] == "self.sss.data"
    assert mod_variables[0]["value"] == "s.read"
    assert mod_variables[1]["value"] == "s.data"

    first_assign = mod_db["scope"][0]
    names = set()
    values = set()
    for entry in first_assign["scope"]:
        if entry["type"] == "assign":
            var = entry["variable"]
            names.add(var["name"])
            values.add(var["value"])

    for i in range(4):
        for j in range(2):
            name = "a.{0}.{1}".format(j, i)
            value = "a[{0}][{1}]".format(j, i)
            assert name in names
            assert value in values


def test_multiple_instance():
    class Mod(Generator):
        def __init__(self, width, param):
            super().__init__("child_{0}_{1}".format(width, param), True)
            in_ = self.input("child_in", width)
            out_ = self.output("child_out", width)

            @always_comb
            def code():
                out_ = 0
                for i in range(param):
                    out_ = in_ + param

            self.add_always(code)

    top = Generator("parent", True)
    top_in = top.input("in", 8)
    top_out = top.output("out", 8)
    children = []
    out_ports = []
    num_instance = 3
    for i in range(num_instance):
        child = Mod(8, i + 2)
        children.append(child)
    for i in range(num_instance):
        child = children[i]
        top.add_child("child_{0}".format(i), child,
                      child_in=top_in)
        out_ports.append(child.ports.child_out)

    top.wire(top_out, reduce_add(*out_ports))

    # testing code
    with tempfile.TemporaryDirectory() as temp:
        filename = os.path.join(temp, "test.sv")
        verilog(top, filename=filename, insert_debug_info=True, debug_fn_ln=True)
        # make sure that the line is mapped correctly
        with open(__file__) as f:
            lines = f.readlines()
            index = -1
            for i in range(len(lines)):
                if "out_ = in_ + param" in lines[i]:
                    index = i
                    break
            assert index != -1
        line_no = index + 1  # 1 offset for line number
        with open(filename + ".debug") as f:
            import json
            content = json.load(f)
        count = 0
        for _, entry in content.items():
            __, ln = entry[0]
            if ln == line_no:
                count += 1
        assert count == num_instance * (num_instance + 3) / 2


def test_ssa_debug():
    mod = Generator("mod", debug=True)
    a = mod.var("a", 4)
    clk = mod.clock("clk")
    b = mod.var("b", 1)
    loop_size = 4

    @always_comb
    def logic1():
        a = 0
        if b:
            for i in range(loop_size):
                a = a + i

    @always_ff((posedge, clk))
    def logic2():
        if a == 1:
            b = 1
        else:
            b = 0   # test_ssa_debug

    mod.add_always(logic1, ssa_transform=True)
    mod.add_always(logic2)

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db, ssa_transform=True)
        with open(debug_db) as f:
            db = json.load(f)
    mod_db = db["table"][0]
    logic1_db = mod_db["scope"][0]["scope"]
    assert len(logic1_db) == (4 + 2 * (loop_size + 1))
    # a_0 = 0
    # 4x (a_i + i)
    idx = 0
    for entry in logic1_db[4:13]:
        var = entry["variable"]
        name = var["name"]
        value = var["value"]
        if name == "i":
            assert value == str(idx)
            idx += 1
        else:
            assert name == "a"
            assert value == "a_{0}".format(idx)
    # enable conditions
    line = get_line_num("b = 0   # test_ssa_debug")
    logic2_db = mod_db["scope"][1]["scope"]
    last_assign = logic2_db[-1]["scope"][-1]["scope"][0]
    assert last_assign["type"] == "assign"
    assert last_assign["condition"] == "!(a == 4'h1)"
    assert last_assign["line"] == line


def test_assign_slice():
    mod = Generator("mod", debug=True)
    a = mod.var("a", 16, size=[2])
    b = mod.input("b", 1)
    c = mod.input("c", 16)
    mod.wire(c, a[b])

    with tempfile.TemporaryDirectory() as temp:
        debug_db = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=debug_db)
        with open(debug_db) as f:
            db = json.load(f)
    mod_db = db["table"][0]
    assert mod_db["name"] == "mod"
    assigns = mod_db["scope"][0]["scope"][3:]
    assert len(assigns) == 2
    for i in range(2):
        assert assigns[i]["variable"]["name"] == "a.{0}".format(i)
        assert assigns[i]["variable"]["value"] == "a[{0}]".format(i)


if __name__ == "__main__":
    test_assign_slice()



import kratos
from kratos import Generator, verilog, const, always_comb, Interface
import platform
import pytest
import os


class PassThroughMod(Generator):
    def __init__(self, is_clone: bool = False):
        super().__init__("mod1", True, is_clone)
        self.in_ = self.input("in", 1)
        self.out_ = self.output("out", 1)
        self.wire(self.out_, self.in_)


def test_regression_1(check_gold):
    class Mod(Generator):
        def __init__(self):
            super().__init__("Top")
            self.in_ = self.input("i_data_in", 2)
            self.out_ = self.output("o_data_out", 2)

            child1 = PassThroughMod()
            self.add_child("child1", child1)
            self.wire(child1.in_, self.in_[0])
            self.wire(self.out_[0], child1.out_)

            child2 = PassThroughMod()
            self.add_child("child2", child2)
            self.wire(self.in_[1], child2.in_)
            self.wire(self.out_[1], child2.out_)

    mod = Mod()
    check_gold(mod, "test_regression_1", optimize_passthrough=False)


def test_regression_2():
    from kratos.func import dpi_function

    @dpi_function(8)
    def func(arg0):
        pass

    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", debug=True)
            self._in = self.input("in", 2)
            self._out = self.output("out", 8)
            self._reg = self.var("val", 8)

            self.add_always(self.code)

        @always_comb
        def code(self):
            # because the types are inferred
            # implicit const conversion doesn't work here
            self._reg = func(const(1, 2))
            self._out = func(const(1, 2))

    mod = Mod()
    mod_src = verilog(mod)
    assert "func" in mod_src[1]


def test_regression_concat():
    from kratos import concat
    parent = Generator("parent")
    for i in range(2):
        child = Generator("child")
        in_ = child.input("in", 1)
        out_ = child.output("out", 1)
        child.wire(out_, in_)
        parent.add_child("child{0}".format(i), child)
    in_ = parent.input("in", 1)
    out_2 = parent.output("out2", 2)
    for i in range(2):
        parent.wire(in_, parent["child{0}".format(i)].ports["in"])
    parent.wire(out_2,
                concat(parent["child0"].ports.out, parent["child1"].ports.out))
    src = verilog(parent, optimize_passthrough=False)["parent"]
    assert "assign out2 = {child0_out, child1_out};" in src


def test_regression_flatten_array_param():
    from _kratos import create_wrapper_flatten
    mod = Generator("mod")
    p = mod.parameter("param", value=16)
    mod.input("in", width=p)
    inst = create_wrapper_flatten(mod.internal_generator, "wrapper")
    gen = Generator("", internal_generator=inst)
    src = verilog(gen)["wrapper"]
    assert "parameter param = 32'h10" in src


def test_regression_interface():
    class ConfigInterface(Interface):
        def __init__(self):
            Interface.__init__(self, "cfg_ifc")
            # Local variables
            wr_en = self.var("wr_en", 1)
            self.master = self.modport("master")
            self.slave = self.modport("slave")
            self.master.set_input(wr_en)
            self.slave.set_input(wr_en)

    class Top(Generator):
        def __init__(self):
            super().__init__("top")
            config = ConfigInterface()
            self.clk = self.clock("clk")
            self.ifc = self.interface(config.slave, "ifc_config_slave", is_port=True)

    verilog(Top())


def test_regression_packed_struct_array():
    from kratos import PackedStruct
    struct = PackedStruct("config_data", [("read", 16, False),
                                          ("data", 16, False)])

    class Parent(Generator):
        def __init__(self):
            super().__init__("parent")
            in_ = self.input("in", 16)
            out = self.output("out", struct, size=2)

            self.wire(out[0]['read'], in_)
            self.wire(out[0]['data'], in_)
            self.wire(out[1]['read'], in_)
            self.wire(out[1]['data'], in_)

    verilog(Parent(), filename="test.sv")


def test_regression_modport_interface():
    from kratos import Interface, verilog, Generator

    class ConfigInterface(Interface):
        def __init__(self):
            super(ConfigInterface, self).__init__("config_i")

            a = self.var("a", 1)
            b = self.var("b", 1)

            m = self.modport("master")
            m.set_output(a)
            self.slave = self.modport("slave")
            self.slave.set_output(b)

    interface = ConfigInterface()
    child = Generator("child")
    child_i = child.interface(interface.slave, "p", is_port=True)
    child.add_stmt(child_i.b.assign(1))
    parent = Generator("parent")
    parent_i = parent.interface(interface.slave, "p", is_port=True)
    parent.add_child("inst", child, p=parent_i)

    v = verilog(parent)["parent"]
    assert ".p(p)" in v


def test_regression_wire_merging():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod")
            self.out_ = self.output("out_", 4)

            self.arr = self.var("arr", 1, size=4)
            self.add_always(self.init)
            self.add_always(self.ctrl)

        @always_comb
        def init(self):
            for i in range(4):
                self.arr[i] = 1

        @always_comb
        def ctrl(self):
            for i in range(4):
                self.out_[i] = self.arr[i]

    m = Mod()
    v = verilog(m)["mod"]
    assert "out_[3] = arr[3];" in v


def test_regression_external_module_wiring():
    module_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "vectors", "module2.sv")

    class Parent(Generator):
        def __init__(self):
            super().__init__(f"parent")
            self.a = self.input("a", 1)
            self.f = self.output("f", 1)
            child = self.from_verilog('child', module_path, [], {})
            self.add_child("child", child)
            self.wire(child.ports['a'], self.a)
            self.wire(child.ports['f'], self.f)

    mod = Parent()
    verilog(mod)


def test_regression_struct_struct_access():
    from kratos import PackedStruct

    class StructTest(Generator):
        def __init__(self):
            super().__init__("struct_test")
            self.wr_tr_struct = PackedStruct("wr_tr_struct", [("wr_en", 1), ("wr_addr", 16)])
            self.rd_tr_struct = PackedStruct("rd_tr_struct", [("rd_en", 1), ("rd_addr", 16)])
            self.tr_struct = PackedStruct("tr_struct")
            self.tr_struct.add_attribute("wr_tr", self.wr_tr_struct)
            self.tr_struct.add_attribute("rd_tr", self.rd_tr_struct)

            self.in_ = self.input("in_", self.tr_struct)
            self.out_ = self.output("out_", 1)

            self.add_always(self.comb)

        @always_comb
        def comb(self):
            if self.in_["wr_tr"]["wr_addr"][4, 0] == 0:
                self.out_ = 1
            else:
                self.out_ = 0

    mod = StructTest()
    verilog(mod)


def test_regression_decouple_wire():
    # test reported from Jake Ke
    from kratos import ternary

    class Child(Generator):
        def __init__(self):
            super().__init__("child")
            self._a = self.input("a", 1)
            self._b = self.output("b", 1)
            self.wire(self._b, 1)

    class Parent(Generator):
        def __init__(self):
            super().__init__("parent")
            self._a_in = self.input("a_in", 1)

            child_inst = Child()
            self.add_child("child_inst", child_inst, a=self._a_in)

            self._c = self.var("c", 1)
            self.add_code(self.some_comb)

            self._d = self.var("d", 1)
            # self.wire(self._d, self._c + self["child_inst"].ports.b) # this is OK for some reason
            # I would like to get the generated wire, child_inst_b
            self.wire(self._d, ternary(self["child_inst"].ports.b, self._c, self._c))
            self._e = self.var("e", 1)
            self.add_always(self.if_comb)
            self._f = self.var("f", 1)
            self.wire(self._f, self["child_inst"].ports.a == 0)

        @always_comb
        def some_comb(self):
            self._c = self["child_inst"].ports.a # I would like to get a_in

        @always_comb
        def if_comb(self):
            if self["child_inst"].ports.a:
                self._e = 1
            else:
                self._e = 0

    mod = Parent()
    src = verilog(mod, fix_port_legality=True)["parent"]
    assert "assign d = child_inst_b ? c: c;" in src
    assert "c = a_in;" in src
    assert "if (a_in) begin" in src
    assert "assign f = a_in == 1'h0;" in src


def test_one_state_fsm():
    # example provided by Max
    class OneStateFSM(Generator):
        def __init__(self, debug: bool = False, is_clone: bool = False, internal_generator=None):
            super().__init__("onestate", debug, is_clone, internal_generator)

            self._clk = self.clock("clk")
            self._rst_n = self.reset("rst_n")

            self.boring_fsm = self.add_fsm("boring_fsm", reset_high=False)
            START = self.boring_fsm.add_state("START")

            self._boring_var = self.var("boring_var", 1)
            START.next(START, None)
            self.boring_fsm.output(self._boring_var)
            START.output(self._boring_var, 1)

            self.boring_fsm.set_start_state(START)

    mod = OneStateFSM()
    verilog(mod)


def test_and():
    mod = Generator("gen", debug=True)
    a = mod.var("a", 4)
    b = mod.var("b", 10)
    c = mod.var("c", 1)

    @always_comb
    def code():
        c = 1
        for i in range(4):
            if (i < a) and (b == const(1, 10)):
                c = 0

    mod.add_code(code)

    src = verilog(mod)["gen"]
    # notice the flipping
    assert "(a > 4'h1) && (b == 10'h1)" in src


def test_static_if():
    mod = Generator("gen")
    a = mod.output("a", 1)
    b = 1

    @always_comb
    def code():
        if b == 1:
            a = 1
        else:
            a = 0

    mod.add_always(code)
    src = verilog(mod)["gen"]
    assert "a = 1'h1" in src


if __name__ == "__main__":
    test_static_if()



import _kratos
import sqlite3
import tempfile
import os
import json
import pytest

from kratos import Generator, Event, always_comb, always_ff, posedge, Transaction, verilog


def test_event_extraction():
    mod = Generator("mod")
    event = Event("test1/event1")
    event_ff = Event("test1/event2")
    t = Transaction("test")

    a = mod.var("a", 1)
    b = mod.var("b", 1)
    c = mod.var("c", 2)
    d = mod.var("d", 2)
    clk = mod.clock("clk")

    @always_comb
    def if_stmt():
        if a:
            t @ event({"a": a})
        else:
            if b:
                t @ event(b=b)
            else:
                t @ event(c=c)

    @always_comb
    def switch_if():
        # this will be turn into a switch statement
        if d == 0:
            t @ event(a1=a)
        elif d == 1:
            t @ event(a2=a)
        elif d == 2:
            t @ event(a3=a)
        else:
            t @ event(a4=a)

    @always_ff((posedge, clk))
    def if_seq():
        if a:
            t @ event_ff(a5=a)

    mod.add_always(if_stmt)
    mod.add_always(switch_if)
    mod.add_always(if_seq)

    # convert to if to switch
    _kratos.passes.transform_if_to_case(mod.internal_generator)
    info = _kratos.extract_event_info(mod.internal_generator)
    assert len(info) == 8
    # check seq
    ffs = [i for i in info if not i.combinational]
    assert len(ffs) == 1
    assert str(ffs[0].fields["a5"]) == "a"
    # check switch
    ffs = [i for i in info if "a4" in i.fields]
    assert len(ffs) == 1
    assert str(ffs[0].fields["a4"]) == "a"
    ffs = [i for i in info if "a3" in i.fields]
    assert len(ffs) == 1
    # check out if statements
    ffs = [i for i in info if "c" in i.fields]
    assert len(ffs) == 1


def test_event_actions():
    mod = Generator("mod")
    event = Event("test1/event1")
    t = Transaction("transaction1")

    a = mod.var("a", 1)
    b = mod.var("b", 1)

    @always_comb
    def code():
        t @ event(a=a).matches(a=a).starts()
        t @ event(b=b).matches(a=b).terminates()

    mod.add_always(code)
    info = _kratos.extract_event_info(mod.internal_generator)
    assert len(info) == 2
    # check actions
    event1 = info[0]
    assert "a" in event1.fields
    stmt = event1.stmt
    assert str(stmt.match_values["a"]) == "a"
    assert event1.transaction == "transaction1"
    assert event1.type == _kratos.EventActionType.Start

    event2 = info[1]
    assert "b" in event2.fields
    stmt = event2.stmt
    assert str(stmt.match_values["a"]) == "b"
    assert event2.type == _kratos.EventActionType.End


def test_event_debug_fn_ln():
    mod = Generator("mod", debug=True)
    event = Event("event")
    t = Transaction("transaction")
    sig = mod.var("sig", 1)

    @always_comb
    def code():
        t @ event(sig=sig)

    mod.add_always(code)
    info = _kratos.extract_event_info(mod.internal_generator)
    stmt = info[0].stmt
    fn_lns = stmt.fn_name_ln
    assert len(fn_lns) == 1
    with open(__file__) as f:
        lines = f.readlines()
    idx = lines.index("        t @ event(sig=sig)\n")
    assert (idx + 1) == fn_lns[0][1]


@pytest.mark.skip("Event currently not working with hgdb")
def test_event_serialization():
    mod = Generator("mod", debug=True)
    event = Event("event")
    t = Transaction("transaction")
    a = mod.var("a", 8)
    b = mod.var("b", 1)
    in_ = mod.input("in", 4)
    out = mod.output("out", 4)
    clk = mod.clock("clk")

    # notice that in kratos we only limit
    @always_ff((posedge, clk))
    def code():
        if a == 0:
            out = in_ + 1
            t @ event(value1=a, value2=b).matches(value2=b).starts()
        elif a == 1:
            out = in_ + 2
            t @ event(value3=a, value4=b).matches(value2=b)
        else:
            out = in_
            t @ event(value1=a, value4=b).matches(value2=b).terminates()

    mod.add_always(code)

    with tempfile.TemporaryDirectory() as temp:
        db_filename = os.path.join(temp, "debug.db")
        verilog(mod, insert_debug_info=True, debug_db_filename=db_filename, contains_event=True)
        conn = sqlite3.connect(db_filename)
        c = conn.cursor()
        c.execute("SELECT * from breakpoint")
        result = c.fetchall()
        # we have 8 lines
        assert len(result) == 8
        event_last = result[-1]
        assert event_last[-2] == "(!(a == 8'h1)) && (!(a == 8'h0))"
        # test out the event table
        c.execute("SELECT * from event")
        result = c.fetchall()
        # 3 events
        assert len(result) == 3
        assert result[0][0] == "event"
        # fields
        fields = json.loads(result[0][3])
        assert fields["value1"] == "mod.a"
        matches = json.loads(result[0][4])
        assert matches["value2"] == "mod.b"
        conn.close()


if __name__ == "__main__":
    test_event_serialization()
